/*
 * Copyright (c) 2009 The openGion Project.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.opengion.hayabusa.taglib;

import static org.opengion.fukurou.util.StringUtil.nval;

import java.io.File;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.sql.Connection;
import java.util.Locale;

import org.opengion.fukurou.business.ArrayTableModel;
import org.opengion.fukurou.business.BizLogic;
import org.opengion.fukurou.db.ConnectionFactory;
import org.opengion.fukurou.util.Closer;
import org.opengion.fukurou.util.ErrMsg;
import org.opengion.fukurou.util.ErrorMessage;
import org.opengion.fukurou.util.HybsLoader;
import org.opengion.fukurou.util.HybsLoaderFactory;
import org.opengion.fukurou.util.HybsLoaderConfig;
import org.opengion.hayabusa.common.HybsSystem;
import org.opengion.hayabusa.common.HybsSystemException;
import org.opengion.hayabusa.db.DBTableModel;

/**
 * 業務ロジックを呼び出すためのタグです。
 * 
 * logics属性に呼び出す業務ロジックのクラス名を記述します。
 * このタグでは、複数の業務ロジックを1度に呼び出すことができ、
 * DB接続のcommit,rollbackは一番最後に、1度のみ実行されます。
 * 各業務ロジックは、記述した順番に呼び出されます。
 * 
 * 業務ロジックは、{@link org.opengion.fukurou.business.BizLogic}インターフェースの
 * 実装クラス、または、この実装クラスを継承したサブクラスである必要があります。
 * 
 * 業務ロジッククラスについては、ホットデプロイ機能により、動的コンパイル、クラスロードが
 * 行われます。
 * 
 * 業務ロジックのソースディレクトリは、システムリソースの BIZLOGIC_SRC_PATH で定義されます。
 * また、同様にクラスディレクトリは、システムリソースの BIZLOGIC_CLASS_PATH で定義されます。
 * さらに、システムリソースの BIZLOGIC_HOTDEPLOY を false に設定することで、動的コンパイル
 * 、クラスロードを行わなくすることもできます。
 * この場合、予めコンパイルされたクラスを、初回呼び出し時に1回のみロードされます。
 *
 * SystemData の USE_SQL_INJECTION_CHECK が true か、quotCheck 属性が true の場合は、
 * ＳＱＬインジェクション対策用のクォーティションチェックを行います。リクエスト引数に
 * クォーティション(')が含まれると、エラーになります。
 * 同様にUSE_XSS_CHECKがtrueか、xssCheck属性がtrueの場合は、
 * クロスサイトススクリプティング(XSS)対策のためless/greater than signのチェックを行います。
 *
 * @og.formSample
 * ●形式：
 *       ・&lt;og:bizLog
 *             logics="業務ロジックのクラス名"
 *             command="ENTRY"
 *             scope="session"
 *             dbid="DEFAULT"
 *             tableId="DEFAULT"
 *             selectedAll="false"
 *             modifyType="A"
 *             keys="SYSTEM_ID"
 *             vals="{&#064;SYSTEM_ID}"
 *             stopError="true"
 *             quotCheck="true"
 *             xssCheck="true"
 *             debug="false"
 *        /&gt;
 * ●body：なし
 *
 * ●使用例
 *     &lt;!-- 業務ロジックの呼び出しを行います --&gt;
 *     &lt;og:bizLogic logics="org.opengion.logic.gf9110.BizLogic_0001" keys="SYSTEM_ID" vals="{&#064;MEM.SYSTEM_ID}" /&gt;
 *
 * @og.rev 5.1.1.0 (2009/12/01) 新規作成
 * @og.group 業務ロジック
 * 
 * @version 5.0
 * @author Hiroki Nakamura
 * @since JDK1.6,
 */
public class BizLogicTag extends CommonTagSupport {
	//* このプログラムのVERSION文字列を設定します。	{@value} */
	private static final String VERSION = "4.0.0 (2005/08/31)" ;
	private static final long serialVersionUID = 4000 ;

	private static final String CMD_ENTRY		= "ENTRY" ;
	private static final String COMMAND_LIST	= CMD_ENTRY;
	private static final String ERR_MSG_ID	 	= HybsSystem.ERR_MSG_KEY;

	private 			String		command		= CMD_ENTRY;
	private				String[]	logics		= null;
	private   			String		dbid		= null ;
	private transient	DBTableModel table		= null;
	private 			String		tableId		= HybsSystem.TBL_MDL_KEY;
	private 			boolean		selectedAll	= false;
	private				String		modifyType	= null;
	private				String[]	keys		= null;
	private				String[]	vals		= null;

	private				boolean		stopError	= true;
	private				boolean		quotCheck	= HybsSystem.sysBool( "USE_SQL_INJECTION_CHECK" );	// 4.0.0 (2005/08/31)
	private				boolean		xssCheck	= HybsSystem.sysBool( "USE_XSS_CHECK" );	// 5.0.0.2 (2009/09/15)

	private transient	ErrorMessage errMessage = null;
	private				int		errCode	 		= ErrorMessage.OK;
	private				int		executeCount	= -1;
	private				Connection		conn	= null;
	private				ArrayTableModel	arrTable= null;
	private				HybsLoader		loader	= null;

	private				String	srcDir			= HybsSystem.sys( "REAL_PATH" ) + HybsSystem.sys( "BIZLOGIC_SRC_PATH" );
	private				String	classDir		= HybsSystem.sys( "REAL_PATH" ) + HybsSystem.sys( "BIZLOGIC_CLASS_PATH" );
	private				boolean	isAutoCompile	= HybsSystem.sysBool( "BIZLOGIC_AUTO_COMPILE" );
	private				boolean	isHotDeploy		= HybsSystem.sysBool( "BIZLOGIC_HOT_DEPLOY" );

	private	static final String	CLASS_PATH;

	// HotDeploy機能を使用する場合に、Javaクラスをコンパイルするためのクラスパスを設定します。
	// 対象となるクラスパスは、WEB-INF/classes 及び WEB-INF/lib/*.jar です。
	static {
		StringBuilder sb = new StringBuilder();
		File lib = new File( HybsSystem.sys( "REAL_PATH" ) + "WEB-INF" + File.separator + "lib" );
		String[] libFiles = lib.list();
		for( int i=0; i<libFiles.length; i++ ) {
			sb.append( libFiles[i] ).append( ';' );
		}
		sb.append( HybsSystem.sys( "REAL_PATH" ) + "WEB-INF" + File.separator + "classes" );

		CLASS_PATH = sb.toString();
	}

	/**
	 * Taglibの開始タグが見つかったときに処理する doStartTag() を オーバーライドします。
	 *
	 * @return	int
	 */
	@Override
	public int doStartTag() {
		if( !check( command, COMMAND_LIST ) ) {
			return(SKIP_BODY);
		}

		startQueryTransaction( tableId );
		return( SKIP_BODY );
	}

	/**
	 * Taglibの終了タグが見つかったときに処理する doEndTag() を オーバーライドします。
	 *
	 * @return	int
	 */
	@Override
	public int doEndTag() {
		debugPrint();

		if( check( command, COMMAND_LIST ) ) {
			useQuotCheck( quotCheck );
			useXssCheck( xssCheck );

			execute();

			String err = TaglibUtil.makeHTMLErrorTable( errMessage,getResource() );
			if( err != null && err.length() > 0 ) {
				setSessionAttribute( ERR_MSG_ID,errMessage );
			}

			if( table != null && ! commitTableObject( tableId, table ) ) {
					jspPrint( "BizLoicTag 処理が割り込まれました。DBTableModel は登録しません。" );
					return (SKIP_PAGE);
			}

			jspPrint( err );
		}

		if( errCode >= ErrorMessage.NG && stopError )  {
			return SKIP_PAGE;
		}
		else {
			return EVAL_PAGE;
		}
	}

	/**
	 * タグリブオブジェクトをリリースします。
	 * キャッシュされて再利用されるので、フィールドの初期設定を行います。
	 */
	@Override
	protected void release2() {
		super.release2();
		command			= CMD_ENTRY;
		logics			= null;
		dbid			= null;
		table 			= null;
		tableId			= HybsSystem.TBL_MDL_KEY;
		selectedAll		= false;
		modifyType      = null;
		keys 			= null;
		vals 			= null;
		stopError		= true;
		quotCheck		= HybsSystem.sysBool( "USE_SQL_INJECTION_CHECK" );
		xssCheck		= HybsSystem.sysBool( "USE_XSS_CHECK" );
		errMessage		= null;
		errCode			= ErrorMessage.OK;
		executeCount	= -1;
		conn			= null;
		arrTable		= null;
		loader			= null;
		srcDir			= HybsSystem.sys( "REAL_PATH" ) + HybsSystem.sys( "BIZLOGIC_SRC_PATH" );
		classDir		= HybsSystem.sys( "REAL_PATH" ) + HybsSystem.sys( "BIZLOGIC_CLASS_PATH" );
		isAutoCompile	= HybsSystem.sysBool( "BIZLOGIC_AUTO_COMPILE" );
		isHotDeploy		= HybsSystem.sysBool( "BIZLOGIC_HOT_DEPLOY" );
	}

	/**
	 * 業務ロジックを実行します。
	 */
	private void execute() {
		int[] rowNos = new int[0];
		table = (DBTableModel)getObject( tableId );

		if( table != null ) {
 			rowNos = getParameterRows();
			String[][] tblVals = new String[rowNos.length][table.getColumnCount()];
			String[] modTypes = new String[rowNos.length];
			for( int i=0; i<rowNos.length; i++ ) {
				tblVals[i] = table.getValues( rowNos[i] );
				modTypes[i] = table.getModifyType( rowNos[i] );
			}
			arrTable = new ArrayTableModel( table.getNames(), tblVals, modTypes );
		}

		conn = ConnectionFactory.connection( dbid, null );
		errMessage = new ErrorMessage();
		loader = HybsLoaderFactory.getLoader(
					new HybsLoaderConfig( srcDir, classDir, isAutoCompile, isHotDeploy, CLASS_PATH )
					);

		boolean rtn = false;
		for( int i=0; i<logics.length; i++ ) {
			BizLogic logic = (BizLogic)loader.newInstance( logics[i] );
			rtn = call( logic );
			if( !rtn ) { break; }
		}

		executeCount = rowNos.length;
		errCode = errMessage.getKekka();
		setRequestAttribute( "DB.COUNT"   , String.valueOf( executeCount ) );
		setRequestAttribute( "DB.ERR_CODE", String.valueOf( errCode ) );

		if( errCode < ErrorMessage.NG ) {
			Closer.commit( conn );

			if( table != null && rowNos.length > 0 ) {
				for( int j=rowNos.length-1; j>=0; j-- ) {
					int row = rowNos[j];
					if( DBTableModel.DELETE_TYPE.equals( table.getModifyType( row ) ) ) {
						table.removeValue( row );
					}
					else {
						table.resetModify( row );
					}
				}
			}
		}
		else {
			Closer.rollback( conn );
		}
		ConnectionFactory.close( conn, dbid );

		// エラーメッセージの行番号を元の選択行に戻します。
		ErrMsg[] errs = errMessage.toArray();
		ErrorMessage errMsgTmp = new ErrorMessage();
		for( int i=0; i<errs.length; i++ ) {
			if( table != null && rowNos.length > 0 ) {
				errMsgTmp.addMessage( errs[i].copy( rowNos[errs[i].getNo()] + 1 ) );
			}
			else {
				errMsgTmp.addMessage( errs[i].copy( errs[i].getNo() + 1 ) );
			}
		}
		errMessage = errMsgTmp;
	}

	/**
	 * 業務ロジックをCallします。
	 * 
	 * @return 業務ロジックの呼び出しが成功したかどうか
	 */
	private boolean call( final BizLogic logic ) {
		if( logic.isRequireTable() ) {
			if( arrTable == null ) { throw new RuntimeException( "TableModelが存在しません。logic=[" + logic.getClass().getSimpleName() + "]" ); }
			else if( arrTable.getRowCount() == 0 ) { return true; } // 0件のときは呼び出ししない
			else { logic.setTable( arrTable ); }
		}
		logic.setConnection( conn );
		logic.setKeys( keys );
		logic.setVals( vals );
		logic.setUserId( getUser().getUserID() );
		logic.setParentPgId( getGUIInfoAttri( "KEY" ) );
		logic.setLoader( loader );
		if( isDebug() ) { logic.setDebug(); }

		boolean rtn = false;
		try {
			rtn = logic.exec();
		}
		catch ( Throwable th ) {
			Closer.rollback( conn );
			ConnectionFactory.remove( conn, dbid );
			throw new HybsSystemException( "業務ロジックの処理中にエラーが発生しました。", th );
		}
		errMessage.append( logic.getErrMsg() );

		setRequestAttribute( logic.getClass().getSimpleName() + ".RETURN"   , logic.getReturn() );
		setRequestAttribute( "RETURN", logic.getReturn() );

		if( isDebug() ) { jspPrint( logic.getDebugMsg() ); }

		return rtn;
	}

	/**\
	 * 【TAG】コマンド(NEW,RENEW)をセットします（ENTRY）。
	 *
	 * @og.tag
	 * コマンドは,HTMLから（get/post)指定されますので,CMD_xxx で設定される
	 * フィールド定数値のいづれかを、指定できます。
	 *
	 * @param	cmd コマンド（public static final 宣言されている文字列)
	 * @see		<a href="{@docRoot}/constant-values.html#org.opengion.hayabusa.taglib.QueryTag.CMD_NEW">コマンド定数</a>
	 */
	public void setCommand( final String cmd ) {
		String cmd2 = getRequestParameter( cmd );
		if( cmd2 != null && cmd2.length() >= 0 ) { command = cmd2.toUpperCase(Locale.JAPAN); }
	}

	/**
	 * 【TAG】実行する業務ロジック名を指定します。
	 *
	 * @og.tag
	 * 実行する業務ロジック名を指定します。業務ロジック名は、クラス名を指定します。
	 * クラス名については、クラス自身の名称のみを指定することができます。
	 * (パッケージ名を含めた完全な形のクラス名を指定することもできます)
	 * また、カンマ区切りで、複数指定することもできます。
	 * この場合、指定した順番に処理されます。
	 * 
	 * @param	lgs 業務ロジック名
	 */
	public void setLogics( final String lgs ) {
		logics = getCSVParameter( lgs );
	}

	/**
	 * 【TAG】(通常は使いません)Queryオブジェクトを作成する時のDB接続IDを指定します。
	 *
	 * @og.tag
	 * Queryオブジェクトを作成する時のDB接続IDを指定します。
	 * これは、システムリソースで、DEFAULT_DB_URL 等で指定している データベース接続先
	 * 情報に、XX_DB_URL を定義することで、 dbid="XX" とすると、この 接続先を使用して
	 * データベースにアクセスできます。
	 *
	 * @param	id データベース接続ID
	 */
	public void setDbid( final String id ) {
		dbid = nval( getRequestParameter( id ),dbid );
	}

	/**
	 * 【TAG】(通常は使いません)結果のDBTableModelを、sessionに登録するときのキーを指定します。
	 *
	 * @og.tag
	 * 検索結果より、DBTableModel オブジェクトを作成します。これを、以下のview 等のタグに
	 * 渡す場合に、通常は、session を利用します。その場合の登録キーです。
	 * query タグを同時に実行して、結果を求める場合、同一メモリに配置される為、
	 * この tableId 属性を利用して、メモリ空間を分けます。
	 * 初期値は、HybsSystem.TBL_MDL_KEY です。
	 *
	 * @param	id sessionに登録する時の ID
	 */
	public void setTableId( final String id ) {
		tableId   = nval( getRequestParameter( id ),tableId );
	}

	/**
	 * 【TAG】データを全件選択済みとして処理するかどうか(true/false)を指定します(初期値:false)。
	 *
	 * @og.tag
	 * 全てのデータを選択済みデータとして扱って処理します。
	 * 全件処理する場合に、（true/false)を指定します。
	 * 初期値は false です。
	 *
	 * @param  all データを全件選択済み(true) / 通常（false)
	 */
	public void setSelectedAll( final String all ) {
		selectedAll = nval( getRequestParameter( all ),selectedAll );
	}

	/**
	 * 【TAG】DB検索時の モディファイタイプ（A(追加),C(更新),D(削除)）を指定します。
	 *
	 * @og.tag
	 * <p>DB検索時に、そのデータをA(追加）、C(更新）、D(削除)のモディファイタイプを
	 * つけた状態にします。<br />
	 * その状態で、そのまま、update する事が可能になります。</p>
	 *
	 * @param   type DB検索時の モディファイタイプ（A,C,D属性）
	 */
	public void setModifyType( final String type ) {
		modifyType = nval( getRequestParameter( type ),modifyType );
	}

	/**
	 * 【TAG】リンク先に渡すキーをCSV形式で複数指定します。
	 *
	 * @og.tag リンク先に渡すキーを指定します。
	 * 分解方法は、CSV変数を先に分解してから、getRequestParameter で値を取得します。
	 * こうしないとデータ自身にカンマを持っている場合に分解をミスる為です。
	 *
	 * @og.rev 3.5.6.2 (2004/07/05) CommonTagSupport#getCSVParameter を使用
	 *
	 * @param	key String
	 */
	public void setKeys( final String key ) {
		keys = getCSVParameter( key );
	}

	/**
	 * 【TAG】リンク先に渡す値をCSV形式で複数指定します。
	 *
	 * @og.tag リンク先に渡す値を指定します。
	 * 分解方法は、CSV変数を先に分解してから、getRequestParameter で値を取得します。
	 * こうしないとデータ自身にカンマを持っている場合に分解をミスる為です。
	 *
	 * @og.rev 3.5.6.2 (2004/07/05) CommonTagSupport#getCSVParameter を使用
	 *
	 * @param	val String
	 */
	public void setVals( final String val ) {
		vals = getCSVParameter( val );
	}
	
	/**
	 * 【TAG】処理エラーの時に処理を中止するかどうかを設定します。
	 *
	 * @og.tag
	 * 初期値は、true(中止する)です。
	 * false(中止しない)に設定する場合、後続処理では、{@DB.ERR_CODE}の値により、
	 * PLSQL/SQLの異常/正常終了によって分岐処理は可能となります。
	 *
	 * @param   flag 中止する ("true")／中止しない("false")
	 */
	public void setStopError( final String flag ) {
		stopError = nval( getRequestParameter( flag ),stopError );
	}

	/**
	 * 【TAG】リクエスト情報の クォーティション(') 存在チェックを実施するかどうか(true/false)を設定します(初期値:USE_SQL_INJECTION_CHECK)。
	 *
	 * @og.tag
	 * ＳＱＬインジェクション対策の一つとして、暫定的ではありますが、SQLのパラメータに
	 * 渡す文字列にクォーティション(') を許さない設定にすれば、ある程度は防止できます。
	 * 数字タイプの引数には、 or 5=5 などのクォーティションを使用しないコードを埋めても、
	 * 数字チェックで検出可能です。文字タイプの場合は、必ず (')をはずして、
	 * ' or 'A' like 'A のような形式になる為、(')チェックだけでも有効です。
	 * (') が含まれていたエラーにする（true)／かノーチェックか(false)を指定します。
	 * 初期値は、SystemData#USE_SQL_INJECTION_CHECK です。
	 *
	 * @param   flag クォーティションチェックする ("true")／しない (それ以外)
	 */
	public void setQuotCheck( final String flag ) {
		quotCheck = nval( getRequestParameter( flag ),quotCheck );
	}

	/**
	 * 【TAG】リクエスト情報の HTMLTag開始/終了文字(&gt;&lt;) 存在チェックを実施するかどうか(true/false)を設定します(初期値:USE_XSS_CHECK)。
	 *
	 * @og.tag
	 * クロスサイトスクリプティング(XSS)対策の一環としてless/greater than signについてのチェックを行います。
	 * (&gt;&lt;) が含まれていたエラーにする（true)／かノーチェックか(false)を指定します。
	 * 初期値は、SystemData#USE_XMLLTAG_CHECK です。
	 *
	 * @param flag boolean XSSチェックする (true)／しない (false)
	 */
	public void setXssCheck( final String flag ) {
		xssCheck = nval( getRequestParameter( flag ),xssCheck );
	}

	/**
	 * 表示データの HybsSystem.ROW_SEL_KEY を元に、選ばれた 行を
	 * 処理の対象とします。
	 *
	 * @return int[]
	 */
	protected int[] getParameterRows() {
		final int[] rowNo ;
		if( selectedAll ) {
			int rowCnt = table.getRowCount();
			rowNo = new int[ rowCnt ];
			for( int i=0; i<rowCnt; i++ ) {
				rowNo[i] = i;
			}
		} else {
			rowNo = super.getParameterRows();
		}
		return rowNo ;
	}
	

	/**
	 * シリアライズ用のカスタムシリアライズ書き込みメソッド
	 *
	 * @serialData
	 *
	 * @param strm ObjectOutputStream
	 */
	private void writeObject( final ObjectOutputStream strm ) throws IOException {
		strm.defaultWriteObject();
	}

	/**
	 * シリアライズ用のカスタムシリアライズ読み込みメソッド
	 *
	 * ここでは、transient 宣言された内部変数の内、初期化が必要なフィールドのみ設定します。
	 *
	 * @serialData
	 *
	 * @param strm ObjectInputStream
	 * @see #release2()
	 */
	private void readObject( final ObjectInputStream strm ) throws IOException , ClassNotFoundException {
		strm.defaultReadObject();
	}

	/**
	 * このオブジェクトの文字列表現を返します。
	 * 基本的にデバッグ目的に使用します。
	 *
	 * @return このクラスの文字列表現
	 */
	public String toString() {
		return org.opengion.fukurou.util.ToString.title( this.getClass().getName() )
				.println( "VERSION"			,VERSION		)
				.println( "command"			,command		)
				.println( "logics"			,logics			)
				.println( "dbid"			,dbid			)
				.println( "tableId"			,tableId		)
				.println( "selectedAll"		,selectedAll	)
				.println( "modifyType"		,modifyType		)
				.println( "keys"			,keys			)
				.println( "vals"			,vals			)
				.println( "stopError"		,stopError		)
				.println( "quotCheck"		,quotCheck		)
				.println( "xssCheck"		,xssCheck		)
				.println( "executeCount"	,executeCount	)
				.println( "errCode"	 		,errCode	 	)
				.println( "Other..."		,getAttributes().getAttribute() )
				.fixForm().toString() ;
	}
}
